#
# Deployment, provisioning and database migration tool
#
# Copyright (C) 2012-2014 TEONITE
# Copyright (C) 2012-2014 Krzysztof Krzysztofik <krzysztof.krzysztofik@teonite.com>
#

from __future__ import print_function

import os
import sys
import logging
import logging.config
import json
import traceback
from collections import Mapping

from fabric.state import env
from fabric.api import run
from fabric.context_managers import contextmanager, prefix

import defaults

_log = None
log = None
config = None


def _setupLogging(config):
    """
    Setup logging
    """
    global _log
    logging.config.dictConfig(config)
    _log = logging


def getLogger(config):
    """
    Get configured logger
    """

    if _log is None:
        if not 'logger' in config:
            _setupLogging(defaults.LOGGING)
        else:
            _setupLogging(config['logger'])

    return logging.getLogger('deployment')


def prepare_logger(config):
    global log
    log = getLogger(config)


def logException(ex):
    log.debug("Error while saving form..")
    exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
    log.error("exception caught while saving form, here's the trace: '%s'",
              repr(traceback.format_exception(exceptionType, exceptionValue, exceptionTraceback))
    )


class NotConfiguredError(Exception):
    pass


def pretty_print(msg, level='debug'):
    if not log:
        print('[%s] ERROR: %s' % (env.host_string, "logger not configured"))
        print('[%s] ERROR: %s' % (env.host_string, msg))
    else:
        if level == 'debug':
            log.debug('[\x1b[36m%s\x1b[0m] \x1b[35mDEBUG\x1b[0m: %s' % (env.host_string, msg))
        elif level == 'info':
            log.info('[\x1b[36m%s\x1b[0m] \x1b[32mINFO\x1b[0m: %s' % (env.host_string, msg))
        elif level == 'error':
            log.error('[\x1b[36m%s\x1b[0m] \x1b[31mERROR\x1b[0m: %s' % (env.host_string, msg))
        elif level == 'warning':
            log.warning('[\x1b[36m%s\x1b[0m] \x1b[33mWARNING\x1b[0m: %s' % (env.host_string, msg))


def _parse_config(filename):
    #pretty_print("Parsing config file: %s" % filename, 'debug')
    try:
        f = open(filename, 'r')
        conf = json.load(f)
        # pretty_print(conf[section], 'debug')

        return conf

    except IOError:
        # exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
        # print("Something went wrong. Returning empty map. Message: %s - %s" % (exceptionType, exceptionValue))
        return {}


def update(d, u, depth=-1):
    """
    Recursively merge or update dict-like objects.
    >>> update({'k1': {'k2': 2}}, {'k1': {'k2': {'k3': 3}}, 'k4': 4})
    {'k1': {'k2': {'k3': 3}}, 'k4': 4}
    """

    for k, v in u.iteritems():
        if isinstance(v, Mapping) and not depth == 0:
            r = update(d.get(k, {}), v, depth=max(depth - 1, -1))
            d[k] = r
        elif isinstance(d, Mapping):
            d[k] = u[k]
        else:
            d = {k: u[k]}
    return d


def prepare_config(config_f=None, remote_tuple=(None, None, None), branch=None):
    global config

    if not config:
        if not config_f:
            config = _parse_config("config.json")
        else:
            config = _parse_config(config_f)

    config = update(defaults.config, config)

    if remote_tuple[0]:
        config['remote']['user'] = remote_tuple[0]
    if remote_tuple[1]:
        config['remote']['host'] = remote_tuple[1]
    if remote_tuple[2]:
        config['remote']['port'] = remote_tuple[2]

    if branch:
        config['source']['git']['branch'] = branch

    return config


def list_dir(dir_=None):
    """returns a list of files in a directory (dir_) as absolute paths"""
    #dir_ = dir_ or env.cwd
    string_ = run("for i in *; do echo $i; done")
    files = string_.replace("\r", "").split("\n")
    return files


def parse_remote(remote_str):
    user = None
    host = None
    port = None

    if "@" in remote_str:
        remote_split = remote_str.rsplit("@", 1)
        user = remote_split[0]
        host = remote_split[1]

        if ":" in user:
            user = user.split(":", 1)[0]
        if ":" in host:
            host_split = host.rsplit(":", 1)

            host = host_split[0]
            port = host_split[1]
    else:
        host = remote_str

        if ":" in host:
            host_split = host.rsplit(":", 1)

            host = host_split[0]
            port = host_split[1]

    return user, host, port


@contextmanager
def source_virtualenv():
    with prefix('source ' + os.path.join(config['venv']['dir'], 'bin/activate')):
        yield



#
# Deployment, provisioning and database migration tool
#
# Copyright (C) 2012-2014 TEONITE
# Copyright (C) 2012-2014 Krzysztof Krzysztofik <krzysztof.krzysztofik@teonite.com>
#


from __future__ import print_function

import shutil
from datetime import datetime

from fabric.context_managers import cd
from fabric.api import put, settings
from fabric.contrib import files
from fabric.context_managers import cd

from git import Repo, InvalidGitRepositoryError

from deployment.libs.gitarchive import GitArchiver
from deployment.common import *
from deployment.plugin import Plugin
from deployment.plugins.info import WriteCommitInfo, NotifyMail
from deployment.plugins.supervisor import Supervisor


class SrcClone(Plugin):
    command = 'src_clone'
    config = None
    description = 'Arguments: <folder> <subfolder> - clone repo to subfolder in local folder'

    def validate_config(self):
        if not 'source' in config:
            raise NotConfiguredError("Source section does not exists")

        if not 'git' in config['source']:
            raise NotConfiguredError("Section \"git\" does not exists")

        if not 'repo' in config['source']['git'] or not len(config['source']['git']['repo']):
            pretty_print("Repository not set. Assuming current dir has a git repo", 'info')
            try:
                g = Repo(os.getcwd())
                config['source']['git']['repo'] = g.remotes.origin.url
            except InvalidGitRepositoryError:
                raise NotConfiguredError("No repo provided, and not in tracked directory")

        if not 'branch' in config['source']['git'] or not len(config['source']['git']['branch']):
            pretty_print("Repository branch not set. Clone from \"master\"", 'info')
            config['source']['git']['branch'] = 'master'

        if not 'local' in config['source']['git']:
            pretty_print("git local repo not set. Using current folder as repo", 'info')
            config['source']['git']['local'] = os.getcwd()

        config['source']['git']['local'] = os.path.expanduser(config['source']['git']['local'])

        if not 'local' in config['source'] or not len(config['source']['local']):
            pretty_print("Local directory not set. Use current working directory", 'info')
            config['source']['local'] = os.getcwd()
        config['source']['local'] = os.path.expanduser(config['source']['local'])

        if not os.path.isdir(config['source']['local']):
            raise NotConfiguredError("config['source']['local'] directory does not exists")

    def run(self, *args, **kwargs):
        self.validate_config()

        try:
            local_directory = args[0]
        except IndexError:
            local_directory = config['source']['local']

        try:
            repo_directory = args[1]
        except IndexError:
            repo_directory = config['source']['git']['local']

        branch = config['source']['git']['branch']
        repo = config['source']['git']['repo']

        env.host_string = 'localhost'
        pretty_print('[+] Repository check start: %s' % local_directory, 'info')

        # if not len(directory):
        #     pretty_print('Directory not selected, assuming current one.', 'info')
        #     directory = os.getcwd()

        # if os.path.isdir(directory):
        #     pretty_print('Directory found.', 'info')
        # else:
            # try:
            #     pretty_print('Directory not found, creating.', 'info')
            #     os.mkdir(directory)
            # except:
            #     raise Exception('Cannot create directory %s, please create the folder manually' % directory)

        old_dir = os.getcwd()
        os.chdir(local_directory)

        pretty_print("Checking if repository exists %s" % os.path.join(local_directory, repo_directory), "info")
        try:
            if not os.path.isdir(os.path.join(repo_directory, ".git")):  # repo = Repo(dir)
                if os.path.isdir(repo_directory):
                    raise IOError('Directory already exists and is not repository. Clone will fail. Please check your '
                                  'configuration')
                raise InvalidGitRepositoryError()

            repo = Repo(repo_directory)

            pretty_print('Repository found.', 'info')

        except InvalidGitRepositoryError:  # Repo doesn't exists
            pretty_print('Repository not found. Creating new one, using %s.' % repo, 'info')
            if len(repo) == 0:
                pretty_print('Repository not selected. Returning.', 'info')
                raise InvalidGitRepositoryError
            repo = Repo.clone_from(repo, repo_directory)

        if not len(branch):
            branch = repo.active_branch

        if repo.active_branch is not branch:
            pretty_print('Changing branch', 'info')
            repo.git.checkout(branch)

        pretty_print('Pulling changes', 'info')
        repo.git.pull('origin', branch)

        pretty_print('Fetching submodules', 'info')
        pretty_print('init')
        repo.git.submodule("init")
        pretty_print('update')
        try:
            repo.submodule_update(init=True, force_remove=True)
        except InvalidGitRepositoryError:
            pretty_print("Cannot update submodules, ommiting", 'info')

        os.chdir(old_dir)
        #repo.create_remote('origin', config.GIT_REPO)

        pretty_print('[+] Repository clone finished', 'info')


class SrcPrepare(Plugin):
    command = 'src_prepare'
    config = None
    description = 'Arguments: <folder> <subfolder> - archive repo from subfolder in local folder to file'

    def validate_config(self):
        if not 'source' in config:
            raise NotConfiguredError("Source section does not exists")

        if not 'git' in config['source']:
            raise NotConfiguredError("Section \"git\" does not exists")

        if not 'dirs' in config['source']['git']:
            pretty_print("No dirs selected, archiving whole repo", "info")
            config['source']['git']['dirs'] = []

        if not 'branch' in config['source']['git'] or not len(config['source']['git']['branch']):
            pretty_print("Repository branch not set. Clone from \"master\"", 'info')
            config['source']['git']['branch'] = 'master'

        if not 'file' in config['source'] or not len(config['source']['file']):
            pretty_print("Archive file not set, using src.tar.gz", 'info')
            config['source']['file'] = 'src.tar.gz'

        config['source']['file'] = os.path.expanduser(config['source']['file'])

    def run(self, *args, **kwargs):
        self.validate_config()

        archive_file = config['source']['file']
        branch = config['source']['git']['branch']
        dirs = config['source']['git']['dirs']
        changelog_file = config['deploy']['changelog']

        try:
            local_directory = os.path.expanduser(args[0])
        except IndexError:
            local_directory = config['source']['local']

        try:
            repo_directory = os.path.expanduser(args[1])
        except IndexError:
            repo_directory = config['source']['git']['local']

        env.host_string = 'localhost'
        pretty_print('[+] Archive prepare start. Branch: %s' % branch, 'info')

        old_dir = os.getcwd()
        os.chdir(local_directory)

        try:
            Repo(repo_directory)
            pretty_print('Repository found.', 'info')
        except InvalidGitRepositoryError:  # Repo doesn't exists
            raise NotConfiguredError('Repository not found. Please provide correct one.')

        pretty_print('Archiving current branch.', 'info')

        # compression = file.split('.')
        pretty_print("cwd: %s" % os.getcwd())

        old_dir2 = os.getcwd()
        os.chdir(repo_directory)

        if len(changelog_file):
            extra = [os.path.join(changelog_file), ]
        else:
            extra = []

        archiver = GitArchiver(main_repo_abspath=os.getcwd(), include_dirs=dirs, extra=extra)
        archiver.create(os.path.join(old_dir2, archive_file))

        os.chdir(old_dir2)
        # if (compression[-1] == "gz" and compression[-2] == "tar") or compression[-1] == "tgz":
        #     repo.git.archive-all('--o', os.path.join(os.getcwd(), file), '--format', "tar.gz", 'HEAD',
        #                          '' if not len(dirs) else dirs)
        # elif compression[-1] == "tar":
        #     repo.git.archive-all('--o', os.path.join(os.getcwd(), file), '--format', "tar", 'HEAD',
        #                          '' if not len(dirs) else dirs)
        # else:
        #     raise Exception("Unknown file format. Supported: tar, tar.gz, tgz")

        os.chdir(old_dir)
        pretty_print('[+] Archive prepare finished', 'info')


class SrcUpload(Plugin):
    command = 'src_upload'
    config = None
    description = 'Arguments: none - upload packed file from local folder to remote host'

    def validate_config(self):
        if not 'local' in config['source'] or not len(config['source']['local']):
            pretty_print("Local directory not set. Use current working directory", 'info')
            config['source']['local'] = os.getcwd()
        config['source']['local'] = os.path.expanduser(config['source']['local'])

        if not 'file' in config['source'] or not len(config['source']['file']):
            pretty_print("Archive file not set, using src.tar.gz", 'info')
            config['source']['file'] = 'src.tar.gz'

        config['source']['file'] = os.path.expanduser(config['source']['file'])

        if not 'remote' in config:
            raise NotConfiguredError("Remote section does not exists")

        if not 'host' in config['remote'] or not len(config['remote']['host']):
            raise NotConfiguredError("Host not set.")

        if not 'user' in config['remote'] or not len(config['remote']['user']):
            raise NotConfiguredError("User not set.")

        if not 'port' in config['remote']:
            pretty_print("Port not set. Using 22", 'info')
            config['remote']['port'] = 22
        env.port = config['remote']['port']

        if not 'dir' in config['remote']:
            pretty_print("Dir not set. Using /tmp/ as default", 'info')
            config['remote']['dir'] = "/tmp/"

    def run(self, *args, **kwargs):
        self.validate_config()

        to_upload = os.path.join(config['source']['local'], config['source']['file'])
        user = config['remote']['user']
        host = config['remote']['host']
        remote_directory = config['remote']['dir']

        env.host = host
        env.user = user
        env.port = config['remote']['port']
        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)

        pretty_print("[+] Starting file '%s' upload (to %s)" % (to_upload, remote_directory), 'info')

        pretty_print('CWD: %s' % os.getcwd())
        old_dir = os.getcwd()

        if not files.exists(remote_directory):
            pretty_print('Target directory not found, creating.', 'info')
            run('mkdir -p %s' % remote_directory)
        else:
            pretty_print('Target directory found, uploading.')

        put(to_upload, "%s" % os.path.join(remote_directory, os.path.basename(to_upload)))
        pretty_print("[+] File upload finished", 'info')
        os.chdir(old_dir)


class SrcLocalClean(Plugin):
    command = 'src_local_clean'
    config = None
    description = 'Arguments: <to_delete> - delete file/folder provided as argument'

    def validate_config(self):
        if not 'local' in config['source'] or not len(config['source']['local']):
            pretty_print("Local directory not set. Use current working directory", 'info')
            config['source']['local'] = os.getcwd()
        config['source']['local'] = os.path.expanduser(config['source']['local'])

    def run(self, *args, **kwargs):
        self.validate_config()

        try:
            to_delete = args[0]
        except IndexError:
            raise NotConfiguredError("src_local_clean need arg 'to_delete' ")
        local_directory = config['source']['local']

        pretty_print('[+] Starting src_local_clean.', 'info')

        old_dir = os.getcwd()
        os.chdir(local_directory)

        if os.path.isfile(to_delete):
            pretty_print('File %s found, deleting' % to_delete, 'info')
            os.remove(to_delete)

        if os.path.isdir(to_delete):
            pretty_print('Directory %s found, deleting.' % to_delete, 'info')
            shutil.rmtree(to_delete)

        os.chdir(old_dir)
        pretty_print('[+] Finished src_local_clean.', 'info')


class SrcRemoteClean(Plugin):
    command = 'src_remote_clean'
    config = None
    description = 'Arguments: <to_delete> - delete file/folder provided as argument'

    def validate_config(self):
        if not 'file' in config['source'] or not len(config['source']['file']):
            raise NotConfiguredError("File name not set.")
        if not 'dir' in config['remote'] or not len(config['remote']['dir']):
            raise NotConfiguredError("Remote dir not set.")

    def run(self, *args, **kwargs):
        self.validate_config()

        try:
            to_delete = args[0]
        except IndexError:
            raise NotConfiguredError("src_remote_clean need arg 'to_delete' ")
        remote_dir = config['remote']['dir']

        pretty_print('[+] Starting src_remote_clean.', 'info')

        remote_path = os.path.join(remote_dir, to_delete)
        run('rm -rf {}'.format(remote_path))
        pretty_print('[+] Finished src_remote_clean.', 'info')


class SrcRemoteCheck(Plugin):
    command = 'src_remote_test'
    config = None
    description = 'Arguments: <user> <host> - test remote host'

    def validate_config(self):
        if not 'remote' in config:
            raise NotConfiguredError("Remote section does not exists")

        if not 'host' in config['remote'] or not len(config['remote']['host']):
            raise NotConfiguredError("Host not set.")

        if not 'user' in config['remote'] or not len(config['remote']['user']):
            raise NotConfiguredError("User not set.")

        if not 'port' in config['remote']:
            pretty_print("Port not set. Using 22", 'info')
            config['remote']['port'] = 22
        env.port = config['remote']['port']

    def run(self, *args, **kwargs):
        try:
            user = args[0]
        except IndexError:
            user = config['remote']['user']

        try:
            host = args[1]
        except IndexError:
            host = config['remote']['host']

        pretty_print("[+] Starting remote test", "info")

        env.host = host
        env.user = user
        env.port = config['remote']['port']
        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)
        #	env.use_ssh_config = True

        run('exit 0')
        pretty_print('[+] Remote test finished', 'info')


class SrcRemoteExtract(Plugin):
    command = 'src_remote_extract'
    config = None
    description = 'Arguments: <subfolder> - extract uploaded file to selected subfolder of deploy_dir, ' \
                  'default current date'

    def validate_config(self):
        if not 'file' in config['source'] or not len(config['source']['file']):
            pretty_print("Archive file not set, using src.tar", 'info')
            config['source']['file'] = 'deployment.tar.gz'

        config['source']['file'] = os.path.expanduser(config['source']['file'])

        if not 'remote' in config:
            raise NotConfiguredError("Remote section does not exists")

        if not 'host' in config['remote'] or not len(config['remote']['host']):
            raise NotConfiguredError("Host not set.")

        if not 'user' in config['remote'] or not len(config['remote']['user']):
            raise NotConfiguredError("User not set.")

        if not 'port' in config['remote']:
            pretty_print("Port not set. Using 22", 'info')
            config['remote']['port'] = 22
        env.port = config['remote']['port']

        if not 'deploy' in config:
            raise NotConfiguredError('No section "deploy" in config.')
        if not 'dir' in config['deploy'] or not len(config['deploy']['dir']):
            config['deploy']['dir'] = 'app'

    def run(self, *args, **kwargs):
        self.validate_config()

        try:
            subfolder = args[0]
        except IndexError:
            subfolder = datetime.now().strftime("%Y%m%d_%H%M%S")

        user = config['remote']['user']
        host = config['remote']['host']
        port = config['remote']['port']
        uploaded_file = config['source']['file']
        file_dir = config['remote']['dir']  # where file is uploaded
        dest_dir = os.path.join(config['deploy']['dir'], subfolder)  # where should be unpacked

        pretty_print("[+] Starting remote extract", 'info')
        pretty_print("Extracting to directory %s" % dest_dir)
        env.host = host
        env.user = user
        env.port = port

        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)
        #	env.use_ssh_config = True

        date = datetime.now().strftime("%Y%m%d-%H%M%S")

        if not files.exists(dest_dir, verbose=True):
            pretty_print('Directory %s does not exists, creating' % dest_dir, 'info')
            run('mkdir -p %s' % dest_dir)
        else:
            pretty_print('Directory %s exists, renaming to %s-%s' % (dest_dir, dest_dir, date), 'info')
            run('mv %s %s-%s' % (dest_dir, dest_dir, date))
            run('mkdir -p %s' % dest_dir)

        #with cd(file_dir):
        pretty_print('Extracting files', 'info')

        compression = uploaded_file.split('.')
        if (compression[-1] == "gz" and compression[-2] == "tar") or compression[-1] == "tgz":
            run('tar xvfz %s -C %s' % (os.path.join(file_dir, uploaded_file), dest_dir))
        # elif compression[-1] == "bz2" and compression[-2] == "tar":
        # 	run('tar xvfj %s -C %s' % (os.path.join(file_dir, file), dest_dir))
        elif compression[-1] == "tar":
            run('tar xvf %s -C %s' % (os.path.join(file_dir, uploaded_file), dest_dir))
        else:
            raise Exception("Unknown file format. Supported: tar, tar.gz, tgz")


class SrcRemoteConfig(Plugin):
    command = 'src_remote_config'
    config = None
    description = 'Arguments: None - copy config from previous to current'

    def validate_config(self):
        if not 'remote' in config:
            raise NotConfiguredError("Remote section does not exists")

        if not 'host' in config['remote'] or not len(config['remote']['host']):
            raise NotConfiguredError("Host not set.")

        if not 'user' in config['remote'] or not len(config['remote']['user']):
            raise NotConfiguredError("User not set.")

        if not 'port' in config['remote']:
            pretty_print("Port not set. Using 22", 'info')
            config['remote']['port'] = 22
        env.port = config['remote']['port']

        if not 'config' in config:
            raise NotConfiguredError('No section "config" in config.')

    def run(self, *args, **kwargs):
        self.validate_config()

        user = config['remote']['user']
        host = config['remote']['host']
        port = config['remote']['port']
        base_dir = config['deploy']['dir']
        to_copy = config['config']

        pretty_print("[+] Starting remote config copy", 'info')

        env.host = host
        env.user = user
        env.port = port
        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)
        #	env.use_ssh_config = True
        if isinstance(to_copy, dict):
            for key, value in to_copy.iteritems():
                (head, tail) = os.path.split(value['dst'])
                if not files.exists(head):
                    pretty_print('Target directory %s does not exists. Creating.' % head, 'info')
                    run('mkdir -p %s' % head)

                if files.exists(value['src'], verbose=True):
                    pretty_print('Copying %s' % key, 'info')
                    run('cp -rf %s %s' % (value['src'], value['dst']))
                else:
                    pretty_print('File does not exists: %s, ommiting' % value['src'], 'error')
        elif isinstance(to_copy, list):
            base_from = os.path.join(base_dir, 'previous')
            base_to = os.path.join(base_dir, 'current')

            for item in to_copy:
                (path, filename) = os.path.split(item)
                from_file = os.path.join(base_from, item)
                to_file = os.path.join(base_to, item)
                to_dir = os.path.join(base_to, path)

                if not files.exists(to_dir):
                    pretty_print('Target directory %s does not exists. Creating.' % path, 'info')
                    run('mkdir -p %s' % to_dir)

                if files.exists(from_file, verbose=True):
                    pretty_print('Copying %s' % filename, 'info')
                    run('cp -rf %s %s' % (from_file, to_file))
                else:
                    pretty_print('File does not exists: %s, ommiting' % filename, 'error')

        else:
            raise NotConfiguredError("Wrong 'config' section format")


class SrcRemoteDeploy(Plugin):
    command = 'src_remote_deploy'
    config = None
    description = 'Arguments: <subfolder> - deploys new version'

    def validate_config(self):
        if not 'remote' in config:
            raise NotConfiguredError("Remote section does not exists")

        if not 'host' in config['remote'] or not len(config['remote']['host']):
            raise NotConfiguredError("Host not set.")

        if not 'user' in config['remote'] or not len(config['remote']['user']):
            raise NotConfiguredError("User not set.")

        if not 'port' in config['remote']:
            pretty_print("Port not set. Using 22", 'info')
            config['remote']['port'] = 22
        env.port = config['remote']['port']

        if not 'deploy' in config:
            raise NotConfiguredError('No section "deploy" in config.')

        if not 'dir' in config['deploy'] or not len(config['deploy']['dir']):
            config['deploy']['dir'] = 'app'

    def run(self, *args, **kwargs):
        self.validate_config()

        try:
            src_dir = args[0]
        except IndexError:
            raise NotConfiguredError('You need to provide folder which will be deployed')

        user = config['remote']['user']
        host = config['remote']['host']
        port = config['remote']['port']
        dst_dir = config['deploy']['dir']

        pretty_print("[+] Starting remote deployment", 'info')

        env.host = host
        env.user = user
        env.port = port
        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)

        with cd(dst_dir):
            with settings(warn_only=True):
                if not run('test -L previous').failed:
                    run('rm -f previous')
                pretty_print("current working dir: %s" % env.cwd)

                if not run('test -L current').failed:
                    run('mv current previous')
            run('ln -s %s current' % os.path.basename(src_dir))

        pretty_print("[+] Remote deployment finished", 'info')


class SrcRemoteRollback(Plugin):
    command = 'src_remote_rollback'
    config = None
    description = 'Arguments: None - backs to previous version'

    def validate_config(self):
        if not 'remote' in config:
            raise NotConfiguredError("Remote section does not exists")

        if not 'host' in config['remote'] or not len(config['remote']['host']):
            raise NotConfiguredError("Host not set.")

        if not 'user' in config['remote'] or not len(config['remote']['user']):
            raise NotConfiguredError("User not set.")

        if not 'port' in config['remote']:
            pretty_print("Port not set. Using 22", 'info')
            config['remote']['port'] = 22
        env.port = config['remote']['port']

        if not 'deploy' in config:
            raise NotConfiguredError('No section "deploy" in config.')

        if not 'dir' in config['deploy'] or not len(config['deploy']['dir']):
            config['deploy']['dir'] = 'app'

    def run(self, *args, **kwargs):
        self.validate_config()

        user = config['remote']['user']
        host = config['remote']['host']
        port = config['remote']['port']
        remote_dir = config['deploy']['dir']

        pretty_print("[+] Starting remote rollback", 'info')

        env.host = host
        env.user = user
        env.port = port
        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)
        #	env.use_ssh_config = True

        with cd(remote_dir):
            with settings(warn_only=True):
                if run('test -L previous').failed:
                    pretty_print('Theres nothing to rollback. Returning.', 'info')
                    return
            run('mv current current.prerollback')
            run('mv previous current')


class SrcPreDeploy(Plugin):
    command = 'src_pre_deploy'
    config = None
    description = 'Arguments: None - runs pre_deploy command from config file'

    def validate_config(self):
        if not 'remote' in config:
            raise NotConfiguredError("Remote section does not exists")

        if not 'host' in config['remote'] or not len(config['remote']['host']):
            raise NotConfiguredError("Host not set.")

        if not 'user' in config['remote'] or not len(config['remote']['user']):
            raise NotConfiguredError("User not set.")

        if not 'port' in config['remote']:
            pretty_print("Port not set. Using 22", 'info')
            config['remote']['port'] = 22
        env.port = config['remote']['port']

        if not 'deploy' in config:
            raise NotConfiguredError('No section "deploy" in config.')

        if not 'pre' in config['deploy'] or not len(config['deploy']['pre']):
            config['deploy']['pre'] = []

    def run(self, *args, **kwargs):
        self.validate_config()

        user = config['remote']['user']
        host = config['remote']['host']
        port = config['remote']['port']
        command_list = config['deploy']['pre']
        remote_dir = config['deploy']['dir']

        env.host = host
        env.user = user
        env.port = port

        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)

        pretty_print("[+] Starting remote pre-deploy commands", 'info')

        if not len(command_list):
            pretty_print('Pre_deploy commands not provided', 'info')
            return

        SrcRemoteCheck(config).run()

        for i in command_list:
            remote_path = os.path.join(remote_dir, 'current')
            if files.exists(remote_path, verbose=True):
                with cd(remote_path):
                    run(i)
            else:
                run(i)

        pretty_print("[+] Remote pre-deploy command finished", 'info')


class SrcPostDeploy(Plugin):
    command = 'src_post_deploy'
    config = None
    description = 'Arguments: None - runs post_deploy command from config file'

    def validate_config(self):
        if not 'remote' in config:
            raise NotConfiguredError("Remote section does not exists")

        if not 'host' in config['remote'] or not len(config['remote']['host']):
            raise NotConfiguredError("Host not set.")

        if not 'user' in config['remote'] or not len(config['remote']['user']):
            raise NotConfiguredError("User not set.")

        if not 'port' in config['remote']:
            pretty_print("Port not set. Using 22", 'info')
            config['remote']['port'] = 22
        env.port = config['remote']['port']

        if not 'deploy' in config:
            raise NotConfiguredError('No section "deploy" in config.')

        if not 'dir' in config['deploy'] or not len(config['deploy']['dir']):
            config['deploy']['dir'] = 'app'

    def run(self, *args, **kwargs):
        self.validate_config()

        user = config['remote']['user']
        host = config['remote']['host']
        port = config['remote']['port']
        command_list = config['deploy']['post']
        remote_dir = config['deploy']['dir']

        env.host = host
        env.user = user
        env.port = port

        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)

        pretty_print("[+] Starting remote post-deploy commands", 'info')

        if not len(command_list):
            pretty_print('Post_deploy commands not provided', 'info')
            return

        SrcRemoteCheck(config).run()

        for i in command_list:
            with cd(os.path.join(remote_dir, 'current')):
                run(i)

        pretty_print("[+] Remote post-deploy command finished", 'info')


class SrcRemoteVenv(Plugin):
    command = 'src_remote_venv'
    config = None
    description = 'Arguments: None - check if venv exists, if not it will be created, and populated with ' \
                  'requirements.txt'

    def validate_config(self):
        if not 'remote' in config:
            raise NotConfiguredError("Remote section does not exists")

        if not 'host' in config['remote'] or not len(config['remote']['host']):
            raise NotConfiguredError("Host not set.")

        if not 'user' in config['remote'] or not len(config['remote']['user']):
            raise NotConfiguredError("User not set.")

        if not 'port' in config['remote']:
            pretty_print("Port not set. Using 22", 'info')
            config['remote']['port'] = 22
        env.port = config['remote']['port']

        if not 'venv' in config:
            raise NotConfiguredError('No section "venv" in config.')

        if not 'dir' in config['venv'] or not len(config['venv']['dir']):
            config['venv']['dir'] = ""

    def run(self, *args, **kwargs):
        self.validate_config()

        check = kwargs.get('check', True)
        install = kwargs.get('install', True)

        user = config['remote']['user']
        host = config['remote']['host']
        port = config['remote']['port']

        try:
            venv_dir = args[0]
        except IndexError:
            venv_dir = config['venv']['dir']

        update = config['venv']['update']
        requirements_files = config['venv']['requirements']

        deploy_dir = config['deploy']['dir']

        if not len(venv_dir):
            pretty_print('[-] Venv dir not set. Returning.')
            return

        update_str = ""
        if update:
            update_str = "U"

        env.host = host
        env.user = user
        env.port = port

        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)

        pretty_print("[+] Starting remote venv check", 'info')

        # SrcRemoteCheck(config).run()
        pretty_print(venv_dir)
        if check:
            if not files.exists(venv_dir):
                pretty_print("[+] Venv not exits, creating", 'info')

                run('virtualenv {} -p /usr/bin/python2'.format(venv_dir))
            else:
                pretty_print("[+] Venv already exists.", 'info')

        if install:
            with source_virtualenv():
                for req_file in requirements_files:
                    req_file = os.path.join(deploy_dir, 'current', req_file)
                    if files.exists("%s" % os.path.join(req_file)):
                        run('pip install -r%s %s' % (update_str, req_file))

        pretty_print("[+] Remote venv check finished", 'info')


class Deploy(Plugin):
    command = 'deploy'
    config = None
    description = 'Arguments: <local_subfolder> - deploys new version'

    def validate_config(self):
        if not 'remote' in config:
            raise NotConfiguredError("Remote section does not exists")

        if not 'clean' in config['source']:
            pretty_print("Local clean flag not set, using True", 'info')
            config['source']['clean'] = True

        if not 'clean' in config['remote']:
            pretty_print("Remote clean flag not set, using True", 'info')
            config['remote']['clean'] = True

    def run(self, *args, **kwargs):
        self.validate_config()

        try:
            subdir = args[0]
        except IndexError:
            subdir = ''

        pretty_print("[+] Starting deployment.", 'info')

        if not config:
            raise NotConfiguredError('Deploy - config not provided')

        date = datetime.now().strftime("%Y%m%d_%H%M%S")

        steps = [SrcRemoteCheck, SrcPreDeploy, SrcClone, WriteCommitInfo, SrcPrepare, SrcUpload, SrcRemoteExtract,
                 SrcRemoteDeploy, SrcRemoteConfig, SrcPostDeploy]

        SrcRemoteVenv(config).run(install=False)

        for step in steps:
            if step in [SrcRemoteExtract, SrcRemoteDeploy]:
                step(config).run(date)
            else:
                step(config).run()

        SrcRemoteVenv(config).run(check=False)

        try:
            Supervisor(config).run()
        except NotConfiguredError:
            pretty_print('Supervisor not configured, ommiting', 'info')

        pretty_print('Local cleaning flag: %s' % config['source']['clean'])
        if config['source']['clean']:
            clean = SrcLocalClean(config)
            clean.run(config['source']['file'])
            clean.run(date)
        else:
            pretty_print('Local cleaning not selected, omitting.', 'info')

        pretty_print('Remote cleaning flag: %s' % config['remote']['clean'])
        if config['remote']['clean']:
            clean = SrcRemoteClean(config)
            clean.run(config['source']['file'])
        else:
            pretty_print('Remote cleaning not selected, omitting.', 'info')

        try:
            NotifyMail(config).run()
        except NotConfiguredError:
            pretty_print('Mail notify not configured, omitting.')

        pretty_print("[+] Deployment finished.", 'info')


#
# Deployment, provisioning and database migration tool
#
# Copyright (C) 2012-2014 TEONITE
# Copyright (C) 2012-2014 Krzysztof Krzysztofik <krzysztof.krzysztofik@teonite.com>
#

__author__ = 'Krzysztof Krzysztofik'



#
# Deployment, provisioning and database migration tool
#
# Copyright (C) 2012-2014 TEONITE
# Copyright (C) 2012-2014 Krzysztof Krzysztofik <krzysztof.krzysztofik@teonite.com>
#

from __future__ import print_function

import xmlrpclib
from fabric.state import env

from deployment.common import NotConfiguredError, pretty_print
from deployment.plugin import Plugin

__author__ = 'kkrzysztofik'


class Supervisor(Plugin):
    command = 'restart_supervisor'
    config = None
    description = 'Arguments: <process_name> <process_name> ... - restart process in Supervisor'

    def validate_config(self):
        #config = prepare_config(config_f, 'Restart')
        pretty_print("Validating supervisor config section", 'debug')

        if not 'supervisor' in self.config:
            raise NotConfiguredError("Supervisor section does not exists")

        supervisor_conf = self.config['supervisor']

        if not 'host' in supervisor_conf or not len(supervisor_conf['host']):
            raise NotConfiguredError("Host not set")

        if not 'port' in supervisor_conf['supervisor']:
            supervisor_conf['port'] = 9001

        if not 'user' in supervisor_conf['supervisor'] or not len(supervisor_conf['user']):
            supervisor_conf['user'] = 'admin'

        if not 'password' in supervisor_conf or not len(supervisor_conf['password']):
            supervisor_conf['password'] = ''

        pretty_print('Config is valid!', 'debug')

    def run(self, *apps, **kwargs):
        if not len(apps):
            if not 'apps' in self.config['supervisor']:
                raise NotConfiguredError("Apps not set")

            if not isinstance(self.config['supervisor']['apps'], list):
                raise Exception("Wrong format of 'apps' section - should be list")

            apps = self.config['supervisor']['apps']

        env.host_string = 'localhost'
        pretty_print('[+] Supervisor restart procedure start', 'info')

        supervisor_conf = self.config['supervisor']
        server = xmlrpclib.Server('http://%s:%s@%s:%i/RPC2' % (
            supervisor_conf['user'], supervisor_conf['password'], supervisor_conf['host'], supervisor_conf['port'])
        )

        env.host_string = "%s:%s" % (supervisor_conf['host'], supervisor_conf['port'])

        for app in apps:
            try:
                pretty_print("Stopping process: %s" % app, "info")
                server.supervisor.stopProcess(app)
                pretty_print("Starting process: %s" % app, "info")
                server.supervisor.startProcess(app)
            except xmlrpclib.Fault as ex:
                if ex.faultCode == 70:
                    pretty_print("Process not running, lets start him", 'info')
                    server.supervisor.startProcess(app)
                else:
                    pretty_print("Something went wrong: %i - %s" % (ex.faultCode, ex.faultString), 'error')
            except xmlrpclib.ProtocolError as prot:
                raise Exception("ProtocolError: %s - %s " % (prot.errcode, prot.errmsg))

        pretty_print('[+] Supervisor restart procedure finished', 'info')



from __future__ import print_function
from __future__ import unicode_literals

__version__ = "1.8"

import logging
from os import extsep, path, readlink
from subprocess import CalledProcessError, Popen, PIPE
import sys
import tarfile
from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED

#taken from https://raw.github.com/Kentzo/git-archive-all/master/git-archive-all
# Copyright (C) 2012 Kentzo
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


class GitArchiver(object):
    """
    GitArchiver

    Scan a git repository and export all tracked files, and submodules.
    Checks for .gitattributes files in each directory and uses 'export-ignore'
    pattern entries for ignore files in the archive.

    >>> archiver = GitArchiver(main_repo_abspath='my/repo/path')
    >>> archiver.create('output.zip')
    """
    LOG = logging.getLogger('GitArchiver')

    def __init__(self, prefix='', exclude=True, force_sub=False, extra=None, main_repo_abspath=None, include_dirs=None):
        """
        @param prefix: Prefix used to prepend all paths in the resulting archive.
            Extra file paths are only prefixed if they are not relative.
            E.g. if prefix is 'foo' and extra is ['bar', '/baz'] the resulting archive will look like this:
            /
              baz
              foo/
                bar
        @type prefix: string

        @param exclude: Determines whether archiver should follow rules specified in .gitattributes files.
        @type exclude:  bool

        @param force_sub: Determines whether submodules are initialized and updated before archiving.
        @type force_sub: bool

        @param extra: List of extra paths to include in the resulting archive.
        @type extra: list

        @param main_repo_abspath: Absolute path to the main repository (or one of subdirectories).
            If given path is path to a subdirectory (but not a submodule directory!) it will be replaced
            with abspath to top-level directory of the repository.
            If None, current cwd is used.
        @type main_repo_abspath: string
        """
        if extra is None:
            extra = []

        if main_repo_abspath is None:
            main_repo_abspath = path.abspath('')
        elif not path.isabs(main_repo_abspath):
            raise ValueError("You MUST pass absolute path to the main git repository.")

        try:
            self.run_shell("[ -d .git ] || git rev-parse --git-dir > /dev/null 2>&1", main_repo_abspath)
        except Exception as e:
            raise ValueError("Not a git repository (or any of the parent directories).".format(path=main_repo_abspath))

        main_repo_abspath = path.abspath(self.read_git_shell('git rev-parse --show-toplevel', main_repo_abspath).rstrip(
        ))

        self.prefix = prefix
        self.exclude = exclude
        self.extra = extra
        self.force_sub = force_sub
        self.main_repo_abspath = main_repo_abspath

        if include_dirs:
            self.include_dirs = " ".join(include_dirs)
        else:
            self.include_dirs = None

    def create(self, output_path, dry_run=False, output_format=None):
        """
        Create the archive at output_file_path.

        Type of the archive is determined either by extension of output_file_path or by output_format.
        Supported formats are: gz, zip, bz2, tar, tgz

        @param output_path: Output file path.
        @type output_path: string

        @param dry_run: Determines whether create should do nothing but print what it would archive.
        @type dry_run: bool

        @param output_format: Determines format of the output archive. If None, format is determined from extension
            of output_file_path.
        @type output_format: string
        """
        if output_format is None:
            file_name, file_ext = path.splitext(output_path)
            output_format = file_ext[len(extsep):].lower()
            self.LOG.debug("Output format is not explicitly set, determined format is {0}.".format(output_format))

        if not dry_run:
            if output_format == 'zip':
                archive = ZipFile(path.abspath(output_path), 'w')

                def add_file(file_path, arcname):
                    if not path.islink(file_path):
                        archive.write(file_path, arcname, ZIP_DEFLATED)
                    else:
                        i = ZipInfo(arcname)
                        i.create_system = 3
                        i.external_attr = 0xA1ED0000
                        archive.writestr(i, readlink(file_path))
            elif output_format in ['tar', 'bz2', 'gz', 'tgz']:
                if output_format == 'tar':
                    t_mode = 'w'
                elif output_format == 'tgz':
                    t_mode = 'w:gz'
                else:
                    t_mode = 'w:{0}'.format(output_format)

                archive = tarfile.open(path.abspath(output_path), t_mode)
                add_file = lambda file_path, arcname: archive.add(file_path, arcname)
            else:
                raise RuntimeError("Unknown format: {0}".format(output_format))

            def archiver(file_path, arcname):
                self.LOG.debug("Compressing {0} => {1}...".format(file_path, arcname))
                add_file(file_path, arcname)
        else:
            archive = None
            archiver = lambda file_path, arcname: self.LOG.info("{0} => {1}".format(file_path, arcname))

        self.archive_all_files(archiver)

        if archive is not None:
            archive.close()

    def get_exclude_patterns(self, repo_abspath, repo_file_paths):
        """
        Returns exclude patterns for a given repo. It looks for .gitattributes files in repo_file_paths.

        Resulting dictionary will contain exclude patterns per path (relative to the repo_abspath).
        E.g. {('.', 'Catalyst', 'Editions', 'Base'), ['Foo*', '*Bar']}

        @type repo_abspath:     string
        @param repo_abspath:    Absolute path to the git repository.

        @type repo_file_paths:  list
        @param repo_file_paths: List of paths relative to the repo_abspath that are under git control.

        @rtype:         dict
        @return:    Dictionary representing exclude patterns.
                    Keys are tuples of strings. Values are lists of strings.
                    Returns None if self.exclude is not set.
        """
        if not self.exclude:
            return None

        def read_attributes(attributes_abspath):
            patterns = []
            if path.isfile(attributes_abspath):
                attributes = open(attributes_abspath, 'r').readlines()
                patterns = []
                for line in attributes:
                    tokens = line.strip().split()
                    if "export-ignore" in tokens[1:]:
                        patterns.append(tokens[0])
            return patterns

        exclude_patterns = {(): []}

        # There may be no gitattributes.
        try:
            global_attributes_abspath = self.read_shell("git config --get core.attributesfile", repo_abspath).rstrip()
            exclude_patterns[()] = read_attributes(global_attributes_abspath)
        except:
            # And it's valid to not have them.
            pass

        for attributes_abspath in [path.join(repo_abspath, f) for f in repo_file_paths if f.endswith(".gitattributes")]:
            # Each .gitattributes affects only files within its directory.
            key = tuple(self.get_path_components(repo_abspath, path.dirname(attributes_abspath)))
            exclude_patterns[key] = read_attributes(attributes_abspath)

        local_attributes_abspath = path.join(repo_abspath, ".git", "info", "attributes")
        key = tuple(self.get_path_components(repo_abspath, repo_abspath))

        if key in exclude_patterns:
            exclude_patterns[key].extend(read_attributes(local_attributes_abspath))
        else:
            exclude_patterns[key] = read_attributes(local_attributes_abspath)

        return exclude_patterns

    def is_file_excluded(self, repo_abspath, repo_file_path, exclude_patterns):
        """
        Checks whether file at a given path is excluded.

        @type repo_abspath: string
        @param repo_abspath: Absolute path to the git repository.

        @type repo_file_path:   string
        @param repo_file_path:  Path to a file within repo_abspath.

        @type exclude_patterns:     dict
        @param exclude_patterns:    Exclude patterns with format specified for get_exclude_patterns.

        @rtype: bool
        @return: True if file should be excluded. Otherwise False.
        """
        if exclude_patterns is None or not len(exclude_patterns):
            return False

        from fnmatch import fnmatch

        file_name = path.basename(repo_file_path)
        components = self.get_path_components(repo_abspath, path.join(repo_abspath, path.dirname(repo_file_path)))

        is_excluded = False
        # We should check all patterns specified in intermediate directories to the given file.
        # At the end we should also check for the global patterns (key '()' or empty tuple).
        while not is_excluded:
            key = tuple(components)
            if key in exclude_patterns:
                patterns = exclude_patterns[key]
                for p in patterns:
                    if fnmatch(file_name, p) or fnmatch(repo_file_path, p):
                        self.LOG.debug("Exclude pattern matched {pattern}: {path}".format(pattern=p,
                                                                                          path=repo_file_path))
                        is_excluded = True

            if not len(components):
                break

            components.pop()

        return is_excluded

    def archive_all_files(self, archiver):
        """
        Archive all files using archiver.

        @param archiver: Function that accepts 2 arguments: abspath to file on the system and relative path within
        archive.
        """
        for file_path in self.extra:
            archiver(path.abspath(file_path), path.join(self.prefix, file_path))

        for file_path in self.walk_git_files():
            archiver(path.join(self.main_repo_abspath, file_path), path.join(self.prefix, file_path))

    def walk_git_files(self, repo_path=''):
        """
        An iterator method that yields a file path relative to main_repo_abspath
        for each file that should be included in the archive.
        Skips those that match the exclusion patterns found in
        any discovered .gitattributes files along the way.

        Recurs into submodules as well.

        @type repo_path:    string
        @param repo_path:   Path to the git submodule repository relative to main_repo_abspath.

        @rtype:     iterator
        @return:    Iterator to traverse files under git control relative to main_repo_abspath.
        """
        repo_abspath = path.join(self.main_repo_abspath, repo_path)
        repo_file_paths = self.read_git_shell("git ls-files --cached --full-name --no-empty-directory %s" % ("" if
                                              not self.include_dirs else self.include_dirs),
                                              repo_abspath).splitlines()
        exclude_patterns = self.get_exclude_patterns(repo_abspath, repo_file_paths)

        for repo_file_path in repo_file_paths:
            # Git puts path in quotes if file path has unicode characters.
            repo_file_path = repo_file_path.strip('"')  # file path relative to current repo
            file_name = path.basename(repo_file_path)

            # Only list symlinks and files that don't start with git.
            if file_name.startswith(".git") or (not path.islink(repo_file_path) and path.isdir(repo_file_path)):
                continue

            main_repo_file_path = path.join(repo_path, repo_file_path)  # file path relative to the main repo

            if self.is_file_excluded(repo_abspath, repo_file_path, exclude_patterns):
                continue

            yield main_repo_file_path

        if self.force_sub:
            self.run_shell("git submodule init", repo_abspath)
            self.run_shell("git submodule update", repo_abspath)

        for submodule_path in self.read_shell("git submodule --quiet foreach 'pwd'", repo_abspath).splitlines():
            # Shell command returns absolute paths to submodules.
            submodule_path = path.relpath(submodule_path, self.main_repo_abspath)
            for file_path in self.walk_git_files(submodule_path):
                yield file_path

    @staticmethod
    def get_path_components(repo_abspath, abspath):
        """
        Split given abspath into components relative to repo_abspath.
        These components are primarily used as unique keys of files and folders within a repository.

        E.g. if repo_abspath is '/Documents/Hobby/ParaView/' and abspath is
        '/Documents/Hobby/ParaView/Catalyst/Editions/Base/', function will return:
        ['.', 'Catalyst', 'Editions', 'Base']

        First element is always '.' (concrete symbol depends on OS).

        @param repo_abspath: Absolute path to the git repository. Normalized via os.path.normpath.
        @type repo_abspath: string

        @param abspath: Absolute path to a file within repo_abspath. Normalized via os.path.normpath.
        @type abspath: string

        @return: List of path components.
        @rtype: list
        """
        repo_abspath = path.normpath(repo_abspath)
        abspath = path.normpath(abspath)

        if not path.isabs(repo_abspath):
            raise ValueError("repo_abspath MUST be absolute path.")

        if not path.isabs(abspath):
            raise ValueError("abspath MUST be absolute path.")

        if not path.commonprefix([repo_abspath, abspath]):
            raise ValueError("abspath (\"{0}\") MUST have common prefix with repo_abspath (\"{1}\")".format(
                abspath, repo_abspath)
            )

        components = []

        while not abspath == repo_abspath:
            abspath, tail = path.split(abspath)

            if tail:
                components.insert(0, tail)

        # Get portable version of '.'
        components.insert(0, path.relpath(repo_abspath, repo_abspath))
        return components

    @staticmethod
    def run_shell(cmd, cwd=None):
        """
        Runs shell command.

        @type cmd:  string
        @param cmd: Command to be executed.

        @type cwd:  string
        @param cwd: Working directory.

        @rtype:     int
        @return:    Return code of the command.

        @raise CalledProcessError:  Raises exception if return code of the command is non-zero.
        """
        p = Popen(cmd, shell=True, cwd=cwd)
        p.wait()

        if p.returncode:
            raise CalledProcessError(returncode=p.returncode, cmd=cmd)

        return p.returncode

    @staticmethod
    def read_shell(cmd, cwd=None, encoding='utf-8'):
        """
        Runs shell command and reads output.

        @type cmd:  string
        @param cmd: Command to be executed.

        @type cwd:  string
        @param cwd: Working directory.

        @type encoding: string
        @param encoding: Encoding used to decode bytes returned by Popen into string.

        @rtype:     string
        @return:    Output of the command.

        @raise CalledProcessError:  Raises exception if return code of the command is non-zero.
        """
        p = Popen(cmd, shell=True, stdout=PIPE, cwd=cwd)
        output, _ = p.communicate()
        output = output.decode(encoding)

        if p.returncode:
            if sys.version_info > (2, 6):
                raise CalledProcessError(returncode=p.returncode, cmd=cmd, output=output)
            else:
                raise CalledProcessError(returncode=p.returncode, cmd=cmd)

        return output

    @staticmethod
    def read_git_shell(cmd, cwd=None):
        """
        Runs git shell command, reads output and decodes it into unicode string

        @type cmd:  string
        @param cmd: Command to be executed.

        @type cwd:  string
        @param cwd: Working directory.

        @rtype:     string
        @return:    Output of the command.

        @raise CalledProcessError:  Raises exception if return code of the command is non-zero.
        """
        p = Popen(cmd, shell=True, stdout=PIPE, cwd=cwd)
        output, _ = p.communicate()
        output = output.decode('unicode_escape').encode('raw_unicode_escape').decode('utf-8')

        if p.returncode:
            if sys.version_info > (2, 6):
                raise CalledProcessError(returncode=p.returncode, cmd=cmd, output=output)
            else:
                raise CalledProcessError(returncode=p.returncode, cmd=cmd)

        return output


#from http://stackoverflow.com/questions/2677617/python-f-write-at-beginning-of-file


class Prepender:
    def __init__(self, fname, mode='w'):
        self.__write_queue = []
        self.__f = open(fname, mode)

    def write(self, s):
        self.__write_queue.insert(0, s)

    def close(self):
        self.__exit__(None, None, None)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.__write_queue: 
            self.__f.writelines(self.__write_queue)
        self.__f.close()


#
# Deployment, provisioning and database migration tool
#
# Copyright (C) 2012-2014 TEONITE
# Copyright (C) 2012-2014 Krzysztof Krzysztofik <krzysztof.krzysztofik@teonite.com>
#

from __future__ import print_function

import sys
import os
from datetime import datetime

from fabric.context_managers import cd, hide
from fabric.operations import put
from fabric.state import env
from fabric.api import run
from fabric.contrib import files

from deployment.common import *
from deployment.plugin import Plugin


class MySQLDumpRemove(Plugin):
    command = 'mysql_dump_remove'
    config = None
    description = 'Arguments: None - remove dump file from remote host'

    def validate_config(self):
        if not 'shell' in config['mysql']:
            raise NotConfiguredError("shell section does not exists")

        if not 'host' in config['mysql']['shell'] or not len(config['mysql']['shell']['host']):
            raise NotConfiguredError("host not set")
        if not 'user' in config['mysql']['shell'] or not len(config['mysql']['shell']['user']):
            raise NotConfiguredError("user not set")
        if not 'port' in config['mysql']['shell'] or not (type(config['mysql']['shell']['port']) == type(int)):
            pretty_print("port not set, using default one", "info")
            config['mysql']['shell']['port'] = 22
            env.port = config['mysql']['shell']['port']

        if not 'dumpfile' in config['mysql']['shell'] or not len(config['mysql']['shell']['dumpfile']):
            pretty_print("migration_dir not set, using default: dump.sql")
            config['mysql']['shell']['migration_dir'] = "dump.sql"

    def run(self, *args, **kwargs):
        self.validate_config()

        user = config['mysql']['shell']['user']
        host = config['mysql']['shell']['host']
        port = config['mysql']['shell']['port']
        filename = config['mysql']['shell']['dumpfile']

        env.host = host
        env.user = user
        env.port = port
        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)

        pretty_print("[+] Starting dump remove.", 'info')

        if files.exists(filename):
            pretty_print('File %s found. Removing.' % filename)
            run('rm %s' % filename)
        else:
            raise Exception('Dump file not found.')

        pretty_print("[+] Dump remove finished.", 'info')


class MySQLDBDump(Plugin):
    command = 'mysql_db_dump'
    config = None
    description = 'Arguments: <database> - dump database to selected file'

    def validate_config(self):
        if not 'server' in config['mysql']:
            raise NotConfiguredError("server section does not exists")

        if not 'host' in config['mysql']['server'] or not len(config['mysql']['server']['host']):
            raise NotConfiguredError("host not set")
        if not 'user' in config['mysql']['server'] or not len(config['mysql']['server']['user']):
            raise NotConfiguredError("user not set")
        if not 'password' in config['mysql']['server'] or not len(config['mysql']['server']['password']):
            raise NotConfiguredError("password not set")
        if not 'port' in config['mysql']['server'] or not (type(config['mysql']['server']['port']) == type(int)):
            pretty_print("port not set, using default one", "info")
            config['mysql']['server']['port'] = 3306
        if not 'database' in config['mysql']['server'] or not len(config['mysql']['server']['database']):
            raise NotConfiguredError("database not set")

        if not 'shell' in config['mysql']:
            raise NotConfiguredError("shell section does not exists")

        if not 'host' in config['mysql']['shell'] or not len(config['mysql']['shell']['host']):
            raise NotConfiguredError("host not set")
        if not 'user' in config['mysql']['shell'] or not len(config['mysql']['shell']['user']):
            raise NotConfiguredError("user not set")
        if not 'port' in config['mysql']['shell'] or not (type(config['mysql']['shell']['port']) == type(int)):
            pretty_print("port not set, using default one", "info")
            config['mysql']['shell']['port'] = 22
            env.port = config['mysql']['shell']['port']

        if not 'dumpfile' in config['mysql']['shell'] or not len(config['mysql']['shell']['dumpfile']):
            pretty_print("migration_dir not set, using default: dump.sql")
            config['mysql']['shell']['migration_dir'] = "dump.sql"

    def run(self, *args, **kwargs):
        self.validate_config()

        try:
            database = args[0]
        except IndexError:
            database = config['mysql']['server']['database']

        user = config['mysql']['shell']['user']
        host = config['mysql']['shell']['host']
        port = config['mysql']['shell']['port']
        filename = config['mysql']['shell']['dumpfile']

        dbhost = config['mysql']['server']['host']
        dbuser = config['mysql']['server']['user']
        dbpassword = config['mysql']['server']['password']

        env.host = host
        env.user = user
        env.port = port
        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)

        pretty_print('[+] Starting MySQL dump.', 'info')

        with hide('running'):
            pretty_print('Running mysqldump.', 'info')
            run('mysqldump -u%s -p%s -h%s %s > %s' % (dbuser, dbpassword, dbhost, database, filename))
        pretty_print('[+] MySQL dump finished.', 'info')


class MySQLDBRestore(Plugin):
    command = 'mysql_db_restore'
    config = None
    description = 'Arguments: <database> - restore database from file'

    def validate_config(self):
        if not 'server' in config['mysql']:
            raise NotConfiguredError("server section does not exists")

        if not 'host' in config['mysql']['server'] or not len(config['mysql']['server']['host']):
            raise NotConfiguredError("host not set")
        if not 'user' in config['mysql']['server'] or not len(config['mysql']['server']['user']):
            raise NotConfiguredError("user not set")
        if not 'password' in config['mysql']['server'] or not len(config['mysql']['server']['password']):
            raise NotConfiguredError("password not set")
        if not 'port' in config['mysql']['server'] or not (type(config['mysql']['server']['port']) == type(int)):
            pretty_print("port not set, using default one", "info")
            config['mysql']['server']['port'] = 3306
        if not 'database' in config['mysql']['server'] or not len(config['mysql']['server']['database']):
            raise NotConfiguredError("database not set")

        if not 'shell' in config['mysql']:
            raise NotConfiguredError("shell section does not exists")

        if not 'host' in config['mysql']['shell'] or not len(config['mysql']['shell']['host']):
            raise NotConfiguredError("host not set")
        if not 'user' in config['mysql']['shell'] or not len(config['mysql']['shell']['user']):
            raise NotConfiguredError("user not set")
        if not 'port' in config['mysql']['shell'] or not (type(config['mysql']['shell']['port']) == type(int)):
            pretty_print("port not set, using default one", "info")
            config['mysql']['shell']['port'] = 22
            env.port = config['mysql']['shell']['port']

        if not 'dumpfile' in config['mysql']['shell'] or not len(config['mysql']['shell']['dumpfile']):
            pretty_print("migration_dir not set, using default: dump.sql")
            config['mysql']['shell']['migration_dir'] = "dump.sql"

    def run(self, *args, **kwargs):
        self.validate_config()

        try:
            database = args[0]
        except IndexError:
            database = config['mysql']['server']['database']

        user = config['mysql']['shell']['user']
        host = config['mysql']['shell']['host']
        port = config['mysql']['shell']['port']
        filename = config['mysql']['shell']['dumpfile']
        dbhost = config['mysql']['server']['host']
        dbuser = config['mysql']['server']['user']
        dbpassword = config['mysql']['server']['password']

        env.host = host
        env.user = user
        env.port = port
        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)

        pretty_print('[+] Starting MySQL restore.', 'info')

        #	env.use_ssh_config = True
        with hide('running'):
            pretty_print('Restoring to %s from %s' % (database, filename), 'info')
            run('mysql -u%s -p%s -h%s %s < %s' % (dbuser, dbpassword, dbhost, database, filename))
        pretty_print('[+] MySQL restore finished.', 'info')


class MySQLDBClone(Plugin):
    command = 'mysql_db_clone'
    config = None
    description = 'Arguments: <db_name> - clone db: <db_name> -> <db_name>_<current_date>_<current_time>'

    def validate_config(self):
        if not 'server' in config['mysql']:
            raise NotConfiguredError("server section does not exists")

        if not 'host' in config['mysql']['server'] or not len(config['mysql']['server']['host']):
            raise NotConfiguredError("host not set")
        if not 'user' in config['mysql']['server'] or not len(config['mysql']['server']['user']):
            raise NotConfiguredError("user not set")
        if not 'password' in config['mysql']['server'] or not len(config['mysql']['server']['password']):
            raise NotConfiguredError("password not set")
        if not 'port' in config['mysql']['server'] or not (type(config['mysql']['server']['port']) == type(int)):
            pretty_print("port not set, using default one", "info")
            config['mysql']['server']['port'] = 3306
        if not 'database' in config['mysql']['server'] or not len(config['mysql']['server']['database']):
            raise NotConfiguredError("database not set")

        if not 'shell' in config['mysql']:
            raise NotConfiguredError("shell section does not exists")

        if not 'host' in config['mysql']['shell'] or not len(config['mysql']['shell']['host']):
            raise NotConfiguredError("host not set")
        if not 'user' in config['mysql']['shell'] or not len(config['mysql']['shell']['user']):
            raise NotConfiguredError("user not set")
        if not 'port' in config['mysql']['shell'] or not (type(config['mysql']['shell']['port']) == type(int)):
            pretty_print("port not set, using default one", "info")
            config['mysql']['shell']['port'] = 22
            env.port = config['mysql']['shell']['port']

        if not 'dumpfile' in config['mysql']['shell'] or not len(config['mysql']['shell']['dumpfile']):
            pretty_print("migration_dir not set, using default: dump.sql")
            config['mysql']['shell']['migration_dir'] = "dump.sql"

    def run(self, *args, **kwargs):
        self.validate_config()

        try:
            database = args[0]
        except IndexError:
            database = config['mysql']['server']['database']

        user = config['mysql']['shell']['user']
        host = config['mysql']['shell']['host']
        port = config['mysql']['shell']['port']
        dbhost = config['mysql']['server']['host']
        dbuser = config['mysql']['server']['user']
        dbpassword = config['mysql']['server']['password']

        env.host = host
        env.user = user
        env.port = port
        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)

        pretty_print('[+] Starting MySQL clone.', 'info')

        new_database = '%s_%s' % (database, datetime.now().strftime("%Y%m%d_%H%M%S"))

        with hide('running'):
            MySQLDBDump(config).run(database)

            pretty_print('Creating new database: %s' % new_database)
            run('mysql -u%s -p%s -h%s %s <<< %s' % (dbuser, dbpassword, dbhost, database,
                                                    '\"CREATE DATABASE %s\"' % new_database))
            MySQLDBRestore(config).run(new_database)

        pretty_print('[+] MySQL clone finished.', 'info')


class MySQLDBMigrate(Plugin):
    command = 'mysql_db_migrate'
    config = None
    description = 'Arguments: <migration_folder> - runs .sql files from selected folder'

    def validate_config(self):
        if not 'server' in config['mysql']:
            raise NotConfiguredError("server section does not exists")

        if not 'host' in config['mysql']['server'] or not len(config['mysql']['server']['host']):
            raise NotConfiguredError("host not set")
        if not 'user' in config['mysql']['server'] or not len(config['mysql']['server']['user']):
            raise NotConfiguredError("user not set")
        if not 'password' in config['mysql']['server'] or not len(config['mysql']['server']['password']):
            raise NotConfiguredError("password not set")
        if not 'port' in config['mysql']['server'] or not (type(config['mysql']['server']['port']) == type(int)):
            pretty_print("port not set, using default one", "info")
            config['mysql']['server']['port'] = 3306
        if not 'database' in config['mysql']['server'] or not len(config['mysql']['server']['database']):
            raise NotConfiguredError("database not set")

        if not 'shell' in config['mysql']:
            raise NotConfiguredError("shell section does not exists")

        if not 'host' in config['mysql']['shell'] or not len(config['mysql']['shell']['host']):
            raise NotConfiguredError("host not set")
        if not 'user' in config['mysql']['shell'] or not len(config['mysql']['shell']['user']):
            raise NotConfiguredError("user not set")
        if not 'port' in config['mysql']['shell'] or not (type(config['mysql']['shell']['port']) == type(int)):
            pretty_print("port not set, using default one", "info")
            config['mysql']['shell']['port'] = 22
            env.port = config['mysql']['shell']['port']

        if not 'dumpfile' in config['mysql']['shell'] or not len(config['mysql']['shell']['dumpfile']):
            pretty_print("migration_dir not set, using default: dump.sql")
            config['mysql']['shell']['migration_dir'] = "dump.sql"

    def run(self, *args, **kwargs):
        self.validate_config()

        try:
            local_dir = args[0]
        except IndexError:
            raise NotConfiguredError("Please provide local migration dir")

        remote_dir = config['mysql']['shell']['migration_dir']
        user = config['mysql']['shell']['user']
        host = config['mysql']['shell']['host']
        port = config['mysql']['shell']['port']
        dbhost = config['mysql']['server']['host']
        dbuser = config['mysql']['server']['user']
        dbpassword = config['mysql']['server']['password']
        database = config['mysql']['server']['database']

        env.user = user
        env.host = host
        env.port = port
        env.host_string = "%s@%s:%s" % (env.user, env.host, env.port)

        pretty_print('[+] Starting MySQL migrate', 'info')
        #	env.use_ssh_config = True

        date = datetime.now().strftime("%Y%m%d_%H%M%S")
        #		pretty_print("Current working directory: %s" % os.getcwd())
        f = []
        for (dirpath, dirname, filenames) in os.walk(local_dir):
            f.extend(filenames)
            break
        f.sort()

        pretty_print('Files: %s' % f)
        os.chdir(local_dir)
        remote_dir = os.path.join(remote_dir, date)
        pretty_print('Creating directory %s' % remote_dir, 'info')
        run('mkdir -p %s' % remote_dir)
        for sql_file in f:
            pretty_print('Uploading file: %s' % sql_file, 'info')
            put(sql_file, remote_dir)

        with cd(remote_dir):
            for sql_file in f:
                pretty_print('Migrating file %s' % sql_file, 'info')
                with hide('running'):
                    run('mysql -u%s -p%s -h%s %s < %s' % (dbuser, dbpassword, dbhost, database, sql_file))
                #			run('rm %s' % file)

        pretty_print('[+] MySQL migrate finished.', 'info')


class DBMigrate(Plugin):
    command = 'db_migrate'
    config = None
    description = 'Arguments: <migration_folder> - migrate database to new version'

    def validate_config(self):
        return True

    def run(self, *args, **kwargs):
        self.validate_config()

        try:
            migration_dir = args[0]
        except IndexError:
            raise NotConfiguredError("Please provide local migration dir")

        pretty_print("[+] Starting database migration.", 'info')

        if not config:
            raise NotConfiguredError

        MySQLDBClone(config).run()
        if migration_dir:
            pretty_print("Migration directory provided. Running.", 'info')
            MySQLDBMigrate(config).run(migration_dir)
        else:
            pretty_print("No migration directory, omitting.", "info")

        MySQLDumpRemove(config).run()

        pretty_print("[+] Database migration finished.", 'info')



#
# Deployment, provisioning and database migration tool
#
# Copyright (C) 2012-2014 TEONITE
# Copyright (C) 2012-2014 Krzysztof Krzysztofik <krzysztof.krzysztofik@teonite.com>
#

__author__ = 'Krzysztof Krzysztofik'

config = {
    "source": {
        "git": {
            "repo": "",
            "branch": "master",
            "local": ".",
            "dirs": []
        },
        "local": ".",
        "file": "src.tar.gz"
    },

    "remote": {
        "host": "",
        "user": "",
        "port": 22,
        "dir": "~",
        "clean": True
    },

    "deploy": {
        "dir": "app",
        "pre": [],
        "post": [],
        "changelog": "",
        "all_commits": True
    },

    "venv": {
        "dir": "",
        "requirements": ["requirements/production.txt"],
        "update": False
    },

    "config": {
    },

    "mysql": {
        "server": {
            "host": "localhost",
            "user": "root",
            "password": "",
            "port": 3306,
            "database": ""
        },
        "shell": {
            "dumpfile": "temp.sql",
            "user": "",
            "host": "",
            "port": 22,
            "migration_dir": ""
        }
    },

    "supervisor": {
        "host": "",
        "port": 9001,
        "user": "admin",
        "password": "",
        "apps": []
    },

    "mail": {
        "project_name": "",

        "from_mail": "",

        "server": {
            "user": "",
            "password": "",
            "host": "localhost",
            "port": "25",
        },

        "people": [
        ],

        "template_path": "",
    }
}

path_list = [
        './config.json',
        './deployment/production.json',
        './deployment/development.json',
        './src/conf/deploy.json',
        './src/conf/deployment.json',
        './src/settings/deploy.json',
        './src/settings/deployment.json',
        './conf/deploy.json',
        './conf/deployment.json',
        './settings/deploy.json',
        './settings/deployment.json',
        './src/conf/deployment/production.json',
        './src/settings/deployment/production.json',
        './conf/deployment/production.json',
        './settings/deployment/production.json'
]

LOGGING = {
    "version": 1,
    "formatters": {
        "simple": {
            "format": "%(message)s"
        },
        "verbose": {
            "format": "[%(asctime)s] \"%(message)s\"",
            "datefmt": "%d/%b/%Y %H:%M:%S"
        }
    },

    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "simple",
            "stream": "ext://sys.stdout"
        }
    },

    "loggers": {
        "root": {
            "handlers": ["console"],
            "level": "INFO"
        },
        "deployment": {
            "handlers": ["console"],
            "level": "INFO",
            "qualname": "deployment",
            "propagate": False
        }
    }
}

SUBJECT_TEMPLATE = "[{project_name}] New version deployed."
MESSAGE_TEMPLATE = "New version of {project_name} has been deployed."


__author__ = 'kkrzysztofik'

import os
import smtplib

from email.MIMEMultipart import MIMEMultipart
from email.MIMEBase import MIMEBase
from email.MIMEText import MIMEText
from email.Utils import COMMASPACE, formatdate
from email import Encoders

from deployment.common import pretty_print


def send_mail(send_from, send_to, subject, text, files=[], host="localhost", user="", password="", port=25):
    assert type(send_to) == list
    assert type(files) == list

    pretty_print('Preparing message')
    msg = MIMEMultipart()
    msg['From'] = send_from
    msg['To'] = COMMASPACE.join(send_to)
    msg['Date'] = formatdate(localtime=True)
    msg['Subject'] = subject

    msg.attach(MIMEText(text))
    pretty_print('Message prepared')

    pretty_print('Preparing files')
    for f in files:
        part = MIMEBase('application', "octet-stream")
        part.set_payload(open(f, "rb").read())
        Encoders.encode_base64(part)
        part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(f))
        msg.attach(part)

    smtp = smtplib.SMTP(host=host, port=port)

    if len(user):
        smtp.starttls()
        smtp.login(user, password)

    smtp.sendmail(send_from, send_to, msg.as_string())
    smtp.close()


from setuptools import setup
from deployment.version import version

readme = []
with open('README.md', 'r') as fh:
    readme = fh.readlines()

setup(
    name='tnt-deployment',
    version=version,
    author='TEONITE',
    author_email='kkrzysztofik@teonite.com',
    packages=['deployment', 'deployment.plugins', 'deployment.libs'],
    scripts=['deploy'],
    url='http://teonite.com/',
    description='Deployment, provisioning and database migration tool.',
    long_description='\n'.join(readme),
    license='GPL 2.0',
    install_requires=[
        "Fabric == 1.9.0",
        "GitPython >= 0.3.2.RC1",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Programming Language :: Python :: 2.7",
        "Topic :: Software Development :: Build Tools",
    ],
    data_files=[('deployment/conf/', ['conf/conf.json.template'])]
)



#
# Deployment and database migration tool
#
# Copyright (C) 2012-2014 TEONITE
# Copyright (C) 2012-2014 Krzysztof Krzysztofik <krzysztof.krzysztofik@teonite.com>
#

from __future__ import print_function

import time

from git import Repo, InvalidGitRepositoryError

from deployment.libs.prepender import Prepender
from deployment.libs.mail import send_mail

from deployment import defaults
from deployment.common import *
from deployment.plugin import Plugin

__author__ = 'kkrzysztofik'


class WriteCommitInfo(Plugin):
    command = 'commit_info'
    config = None
    description = 'Arguments: <none> - write to changelog file info about commit since last build'

    def validate_config(self):
        pass

    def run(self, *args, **kwargs):
        changelog_filename = config['deploy']['changelog']
        all_commits = config['deploy']['all_commits']
        local_directory = config['source']['local']
        repo_directory = config['source']['git']['local']
        branch = config['source']['git']['branch']

        env.host_string = 'localhost'
        pretty_print('[+] Write commit info start', 'info')

        if not len(changelog_filename):
            pretty_print("No changelog file provided, returning", 'info')
            return

        old_dir = os.getcwd()
        os.chdir(local_directory)

        try:
            with open("previous_build", "r") as build_file:
                last_build = build_file.readline().strip()
        except IOError:
            try:
                last_build = os.environ['GIT_PREVIOUS_COMMIT']
            except KeyError:
                last_build = None

        try:
            repo = Repo(repo_directory)
            pretty_print('Repository found.', 'info')
        except InvalidGitRepositoryError:  # Repo doesn't exists
            raise NotConfiguredError('Repository not found. Please provide correct one.')

        if not len(branch):
            branch = repo.active_branch

        if repo.active_branch is not branch:
            pretty_print('Changing branch', 'info')
            repo.git.checkout('master')

        pretty_print('Pulling changes', 'info')
        repo.git.pull('origin', branch)

        last_commit = repo.head.commit

        if not all_commits:
            mode = "a"
            if last_build:
                commits = "%s..HEAD" % last_build
            else:
                commits = "-n1"
        else:
            mode = "w"
            commits = None

        with Prepender(os.path.join(local_directory, repo_directory, changelog_filename), mode) as changelog_file:
            head_str = u'%s, %s - %s <%s>\n' % (last_commit.hexsha[:8],
                                                time.strftime(u"%a, %d %b %Y %H:%M", time.gmtime(last_commit.committed_date)),
                                                last_commit.committer.name, last_commit.committer.email)

            pretty_print("Getting git log", "info")
            if commits:
                gitlog = repo.git.log(commits, u'--format=* [%h] - %s')
            else:
                gitlog = repo.git.log(u'--format=* [%h] - %s')

            if len(gitlog):
                pretty_print("Formatting gitlog", "info")
                write_str = head_str + u"%s\n\n" % (u"#" * len(head_str)) + unicode(gitlog, 'utf8') + u"\n\n"
                pretty_print("Writting gitlog to file", "info")
                changelog_file.write(write_str.encode("utf8"))
                pretty_print("Git log written")
            else:
                pretty_print("No git log to write", "info")

        # compression = file.split('.')
        pretty_print("cwd: %s" % os.getcwd())

        if not all_commits:
            with open("previous_build", "w") as build_file:
                build_file.write(last_commit.hexsha)

        os.chdir(old_dir)
        pretty_print('[+] Write commit info finished', 'info')


class NotifyMail(Plugin):
    command = 'notify'
    config = None
    description = 'Arguments: <none> - send mail with deploy notification'

    def validate_config(self):
        if not len(config['mail']['project_name']):
            raise NotConfiguredError("Project name not set.")

        if not len(config['mail']['from_mail']):
            raise NotConfiguredError("From: mail not set.")

        if not len(config['mail']["people"]):
            raise NotConfiguredError("No people to notify")

    def run(self, *args, **kwargs):
        self.validate_config()

        project_name = config['mail']['project_name']
        from_mail = config['mail']['from_mail']
        people = config['mail']["people"]
        host = config['mail']['server']['host']
        user = config['mail']['server']['user']
        password = config['mail']['server']['password']
        port = config['mail']['server']['port']
        local_dir = config['source']['local']
        repo_dir = config['source']['git']['local']

        template_path = os.path.join(local_dir, repo_dir, config['mail']['template_path'])

        env.host_string = 'localhost'
        pretty_print('[+] Notify mail start', 'info')

        if not len(template_path):
            subject = defaults.SUBJECT_TEMPLATE.format(project_name=project_name)
            message = defaults.MESSAGE_TEMPLATE.format(project_name=project_name)
        else:
            template_path = os.path.expanduser(template_path)
            head, tail = os.path.split(template_path)

            if tail.endswith(".py"):
                tail = tail[:-3]
            else:
                raise NotConfiguredError("Template file must be .py file")

            sys.path.insert(0, head)
            templates = __import__(tail)

            subject = templates.SUBJECT_TEMPLATE.format(project_name=project_name)
            message = templates.MESSAGE_TEMPLATE.format(project_name=project_name)

        message = "".join([message, "\n\n--\nSent automatically by TEONITE Deployment (http://teonite.com)\nCopyright (C) TEONITE (http://teonite.com)"])
        send_mail(from_mail, people, subject, message, host=host, user=user, password=password, port=port)

        pretty_print('[+] Notify mail finished', 'info')



#
# Deployment, provisioning and database migration tool
#
# Copyright (C) 2012-2014 TEONITE
# Copyright (C) 2012-2014 Krzysztof Krzysztofik <krzysztof.krzysztofik@teonite.com>
#

__author__ = 'kkrzysztofik'

import os
import sys


class Plugin(object):
    command = None
    config = None
    description = None

    def __init__(self, config, **kwargs):
        self.config = config

    def validate_config(self):
        raise NotImplementedError()

    def run(self, *args, **kwargs):
        self.validate_config()
        raise NotImplementedError()


def init_plugins(config):
    find_plugins()
    plugins = register_plugins(config)

    return plugins


def find_plugins():
    """
        find all files in the plugin directory and imports them
    """
    plugin_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "plugins")
    # print plugin_dir
    plugin_files = [x[:-3] for x in os.listdir(plugin_dir) if x.endswith(".py")]
    sys.path.insert(0, plugin_dir)

    for plugin in plugin_files:
        mod = __import__(plugin)


def register_plugins(config):
    """
        Register all class based plugins.

        Uses the fact that a class knows about all of its subclasses
        to automatically initialize the relevant plugins
    """
    plugins = {}

    for plugin in Plugin.__subclasses__():
        plugins[plugin.command] = plugin(config)

    return plugins
